import importlib
import json
import requests
from typing import Optional
from docassemble.base.util import log, get_config, interview_url

# reference: https://gist.github.com/JeffPaine/3145490
# https://docs.github.com/en/free-pro-team@latest/rest/reference/issues#create-an-issue

# Authentication for user filing issue (must have read/write access to
# repository to add issue to)
__all__ = ["valid_github_issue_config", "make_github_issue", "feedback_link"]
USERNAME = get_config("github issues", {}).get("username")
TOKEN = get_config("github issues", {}).get("token")


def valid_github_issue_config():
    return bool(TOKEN)


def feedback_link(
    user_info_object=None,
    i: str = None,
    github_repo: str = None,
    github_user: str = None,
    variable: str = None,
    question_id: str = None,
    package_version: str = None,
    filename: str = None,
) -> str:
    """
    Helper function to get a link to the GitHub feedback form.
    For simple usage, it should be enough to call
    `feedback_link(user_info(), github_user="MY_USER_OR_ORG_ON_GITHUB")`, so long as the package you
    want to provide feedback on exists on GitHub and you are running from an "installed" (not playground)
    link.
    """
    if user_info_object:
        package_name = str(user_info_object.package)
        # TODO: we can use the packages table or /api/packages to get the exact GitHub URL
        if package_name and not package_name.startswith("docassemble-playground"):
            _github_repo = package_name.replace(".", "-")
        else:
            _github_repo = "demo"  # default repo on GitHub of suffolklitlab-issues/demo for documentation purposes
        _variable = user_info_object.variable
        _question_id = user_info_object.question_id
        _filename = user_info_object.filename
        try:
            _package_version = str(
                importlib.import_module(user_info_object.package).__version__
            )
        except:
            _package_version = "playground"
        if get_config("github issues", {}).get("default repository owner"):
            _github_user = get_config("github issues", {}).get(
                "default repository owner"
            )
        _session_id = user_info_object.session

    # Allow keyword params to override any info from the user_info() object
    # We will try pulling the repo owner name from the Docassemble config
    if github_repo and github_user:
        _github_repo = github_repo
        _github_user = github_user
    elif (
        get_config("github issues", {}).get("default repository owner") and github_repo
    ):
        _github_user = get_config("github issues", {}).get("default repository owner")
        _github_repo = github_repo
    else:
        _github_repo = "demo"
        _github_user = "suffolklitlab-issues"
    if variable:
        _variable = variable
    if question_id:
        _question_id = question_id
    if package_version:
        _package_version = package_version
    if filename:
        _filename = filename

    if not i:
        i = "docassemble.GithubFeedbackForm:feedback.yml"

    return interview_url(
        i=i,
        github_repo=_github_repo,
        github_user=_github_user,
        variable=_variable,
        question_id=_question_id,
        package_version=_package_version,
        filename=_filename,
        session_id=_session_id,
        local=False,
        reset=1,
    )


def make_github_issue(
    repo_owner, repo_name, template=None, title=None, body=None, label=None
) -> Optional[str]:
    """
    Create a new Github issue and return the URL.

    template - the docassemble template for the github issue. Overrides `title` and `body` if provided.
    title - the title for the github issue
    body - the body of the github issue
    """
    make_issue_url = f"https://api.github.com/repos/{repo_owner}/{repo_name}/issues"
    # Headers
    if not TOKEN:
        log("Error creating issues: No valid GitHub token provided.")
        return None

    headers = {
        "Authorization": "token %s" % TOKEN,
        "Accept": "application/vnd.github.v3+json",
    }

    if label:
        labels_url = f"https://api.github.com/repos/{repo_owner}/{repo_name}/labels"
        has_label_resp = requests.get(labels_url + "/" + label, headers=headers)
        if has_label_resp.status_code == 404:
            label_data = {
                "name": label,
                "description": "Feedback from a Docassemble Interview",
                "color": "002E60",
            }
            make_label_resp = requests.post(
                labels_url, data=json.dumps(label_data), headers=headers
            )
            if make_label_resp.status_code != 201:
                log(
                    f"Was not able to find nor create the {label} label: {make_label_resp.text}"
                )
                label = None

    if template:
        title = template.subject
        body = template.content
    # Create our issue
    data = {
        "title": title,
        "body": body,
    }
    if label:
        data["labels"] = [label]

    # Add the issue to our repository
    response = requests.post(make_issue_url, data=json.dumps(data), headers=headers)
    if response.status_code == 201:
        return response.json().get("html_url")
    else:
        log(f'Could not create Issue "{title}", results {response.text}')
        return None



import os
import sys
from setuptools import setup, find_packages
from fnmatch import fnmatchcase
from distutils.util import convert_path

standard_exclude = ('*.pyc', '*~', '.*', '*.bak', '*.swp*')
standard_exclude_directories = ('.*', 'CVS', '_darcs', './build', './dist', 'EGG-INFO', '*.egg-info')
def find_package_data(where='.', package='', exclude=standard_exclude, exclude_directories=standard_exclude_directories):
    out = {}
    stack = [(convert_path(where), '', package)]
    while stack:
        where, prefix, package = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if os.path.isdir(fn):
                bad_name = False
                for pattern in exclude_directories:
                    if (fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()):
                        bad_name = True
                        break
                if bad_name:
                    continue
                if os.path.isfile(os.path.join(fn, '__init__.py')):
                    if not package:
                        new_package = name
                    else:
                        new_package = package + '.' + name
                        stack.append((fn, '', new_package))
                else:
                    stack.append((fn, prefix + name + '/', package))
            else:
                bad_name = False
                for pattern in exclude:
                    if (fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()):
                        bad_name = True
                        break
                if bad_name:
                    continue
                out.setdefault(package, []).append(prefix+name)
    return out

setup(name='docassemble.GithubFeedbackForm',
      version='0.2.1',
      description=('Integration of GitHub issues with Docassemble interviews for collecting user feedback'),
      long_description='# docassemble.GithubFeedbackForm\r\n\r\nA package that uses the GitHub API to gather feedback and then submit issues to Github that can be embedded\r\ninto a Docassemble interview. Makes it easy to collect per-page feedback.\r\n\r\nThis package is designed to support the following workflow:\r\n\r\n1. Work is stored on a public GitHub repository, or at least, you setup a repository to collect feedback.\r\n1. There is one package per "interview"/"app".\r\n2. Each question block has a unique question ID.\r\n3. Preferably--questions are triggered in an interview order block. If you use a series of `mandatory`\r\n  blocks instead of a single mandatory block, the `variable` listed in the bug report may not be as useful.\r\n\r\n## Getting started\r\n\r\n1. Create a new GitHub user and create a personal access token on it. The personal access\r\n   token needs minimal permissions. Specifically, it needs to be allowed to make pull requests.\r\n   Pull request access is allowed for anyone by default when you create a new, public GitHub repository.\r\n3. Edit your config, and create a block like this:\r\n\r\n```yaml\r\ngithub issues:\r\n  username: "YOUR_NEW_DEDICATED_ISSUE_CREATION_ACCOUNT"\r\n  token: "..." # A valid GitHub personal access token associated with the username above\r\n  default repository owner: YOUR_GITHUB_USER_OR_ORG_HERE\r\n  allowed repository owners: # List the repo that your account will be allowed to create issues on\r\n    - YOUR_GITHUB_USER_OR_ORG_HERE \r\n    - SECOND_GITHUB_USER_OR_ORG\r\n```\r\n\r\n  Note that it is important to provide a list of allowed repository owners.\r\n  This is used to prevent your form from being used to spam GitHub \r\n  repositories with feedback.\r\n  \r\n3. Add a link on each page, in the footer or `under` area.  \r\n   You can use the `feedback_link()` function to add a link, like this:\r\n   `[:comment-dots: Feedback](${ feedback_link(user_info()) } ){:target="_blank"}`\r\n   \r\n   Optional parameters:\r\n    - `i`: the feedback form, like: docassemble.AssemblyLine:feedback.yml\r\n    - `github_repo`: repo name, like: docassemble-AssemblyLine\r\n    - `github_user`: owner of the repo, like: suffolklitlab\r\n    - `variable`: variable being sought, like: intro\r\n    - `question_id`:  id of the current question, like: intro\r\n    - `package_version`: version number of the current package\r\n    - `filename`: filename of the interview the user is providing feedback on.\r\n    \r\n    Each has a sensible default. Most likely, you will limit your custom\r\n    parameters to the `github_repo` if you want feedback links to work\r\n    from the docassemble playground.\r\n    \r\n    You will also need to include the `github_issue.py` module in your parent interview,\r\n    like this: \r\n    ```yaml\r\n    ---\r\n    modules:\r\n      - docassemble.GithubFeedbackForm.github_issue\r\n    ```\r\n    \r\n4. Optionally, create your own feedback.yml file. If you want a custom feedback.yml,\r\n   it should look like this, with whatever customizations you choose:\r\n\r\n```yaml\r\ninclude:\r\n  - docassemble.GithubFeedbackForm:feedback.yml\r\n---\r\ncode: |\r\n  al_feedback_form_title = "Your title here"  \r\n---\r\ncode: |\r\n  # This email will be used ONLY if there is no valid GitHub config\r\n  al_error_email = "your_email@yourdomain.com"\r\n---\r\ntemplate: al_how_to_get_legal_help\r\ncontent: |\r\n  If you need more help, these are free resources:\r\n\r\n  ... [INCLUDE STATE-SPECIFIC RESOURCES]      \r\n```\r\n\r\nYou may also want to customize the metadata: title, exit url and override \r\nany specific questions, add a logo, etc.    \r\n\r\n## Author\r\n\r\nQuinten Steenhuis, qsteenhuis@suffolk.edu\r\n',
      long_description_content_type='text/markdown',
      author='Quinten Steenhuis',
      author_email='qsteenhuis@suffolk.edu',
      license='The MIT License (MIT)',
      url='https://courtformsonline.org',
      packages=find_packages(),
      namespace_packages=['docassemble'],
      install_requires=['docassemble.ALToolbox>=0.6.0'],
      zip_safe=False,
      package_data=find_package_data(where='docassemble/GithubFeedbackForm/', package='docassemble.GithubFeedbackForm'),
     )




import importlib

from typing import Optional, Iterable, Tuple
from datetime import datetime
from sqlalchemy import (
    insert,
    update,
    select,
    Text,
    DateTime,
    Table,
    Column,
    String,
    Integer,
    MetaData,
    create_engine,
    func,
)
from docassemble.base.util import DARedis, log
from docassemble.base.sql import alchemy_url

__all__ = [
    "save_feedback_info",
    "set_feedback_github_url",
    "redis_panel_emails_key",
    "add_panel_participant",
    "potential_panelists",
    "get_all_feedback_info",
    "save_good_or_bad",
    "get_good_or_bad",
]

redis_panel_emails_key = "docassemble-GithubFeedbackForm:panel_emails"

############################################
## Panel particitpants are managed through a Redis ZSet (a sorted/scored set).
## The score is the timestamp of when they responded, so we can fetch newer
## respondents (more likely to accept a follow up interview).
## https://redis.io/docs/manual/data-types/#sorted-sets


def add_panel_participant(email: str):
    """Adds this email to a list of potential participants for a qualitative research panel.
    Adds the email and when they responded, as we shouldn't link their feedback to their identity at all.
    """
    red = DARedis()
    red.zadd(redis_panel_emails_key, {email: datetime.now().timestamp()})


def potential_panelists() -> Iterable[Tuple[str, datetime]]:
    red = DARedis()
    return [
        (item, datetime.fromtimestamp(score))
        for item, score in red.zrange(redis_panel_emails_key, 0, -1, withscores=True)
    ]


###################################
## Using SQLAlchemy to save / retrieve session information that is linked
## to specific feedback issues

db_url = alchemy_url("db")
engine = create_engine(db_url)

metadata_obj = MetaData()

## This table functions as both storage for bug report-session links,
## and for more general on-server feedback, prompted for at the end of interviews
feedback_session_table = Table(
    "feedback_session",
    metadata_obj,
    Column("id", Integer, primary_key=True),
    Column("interview", String),
    Column("session_id", String),
    Column("body", Text),
    Column("html_url", String),
)

good_or_bad_table = Table(
    "good_or_bad",
    metadata_obj,
    Column("reaction", Integer),
    Column("interview", String),
    Column("version", String),
    Column("datetime", DateTime),
)

metadata_obj.create_all(engine)


def save_feedback_info(
    interview, *, session_id=None, template=None, body=None
) -> Optional[str]:
    """Saves feedback along with optional session information in a SQL DB"""
    if template:
        body = template.content

    if interview and (session_id or body):
        stmt = insert(feedback_session_table).values(
            interview=interview, session_id=session_id, body=body
        )
        with engine.begin() as conn:
            result = conn.execute(stmt)
            id_for_feedback = result.inserted_primary_key[0]

        return id_for_feedback
    else:  # can happen if the forwarding interview didn't pass session info
        return None


def set_feedback_github_url(id_for_feedback: str, github_url: str) -> bool:
    """Returns true if save was successful"""
    stmt = (
        update(feedback_session_table)
        .where(feedback_session_table.c.id == id_for_feedback)
        .values(html_url=github_url)
    )
    with engine.begin() as conn:
        result = conn.execute(stmt)
    if result.rowcount == 0:
        log(f"Cannot find {id_for_feedback} in redis DB")
        return False
    return True


def get_all_feedback_info(interview=None) -> Iterable:
    if interview:
        stmt = select(feedback_session_table).where(
            feedback_session_table.c.interview == interview
        )
    else:
        stmt = select(feedback_session_table)
    with engine.begin() as conn:
        results = conn.execute(stmt)
        # Turn into literal dict because DA is too eager to save / load SQLAlchemy objects into the interview SQL
        return {str(row["id"]): dict(row) for row in results.mappings()}


def save_good_or_bad(
    reaction: int, *, user_info_object=None, interview: str = None, version: str = None
):
    """Saves a user's reaction to an interview, in the form of an int (0 being
    neutral, positive numbers being good, and negative numbers being bad)"""
    _package_version = None
    _interview = None
    if user_info_object:
        _interview = user_info_object.filename
        try:
            _package_version = str(
                importlib.import_module(user_info_object.package).__verison__
            )
        except:
            _package_version = "playground"

    if interview:
        _interview = interview
    if version:
        _package_version = version

    stmt = insert(good_or_bad_table).values(
        reaction=reaction,
        interview=_interview,
        version=_package_version,
        datetime=datetime.now(),
    )
    with engine.begin() as conn:
        conn.execute(stmt)


def get_good_or_bad(interview: str = None):
    """Retrieves user's aggregate reactions to an interview (how many reactions
    and the average score of them), grouped by interview and package verison"""
    stmt = select(
        good_or_bad_table.c.interview,
        good_or_bad_table.c.version,
        func.count().label("count"),
        func.avg(good_or_bad_table.c.reaction).label("average"),
    )
    if interview:
        stmt = stmt.where(good_or_bad_table.c.interview == interview)
    stmt = stmt.group_by(good_or_bad_table.c.interview, good_or_bad_table.c.version)
    with engine.begin() as conn:
        results = conn.execute(stmt)
        return [dict(row) for row in results.mappings()]
